@@ -16,7 +16,7 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
-#define _GNU_SOURCE
+#define _LGPL_SOURCE
 #include <fcntl.h>
 #include <getopt.h>
 #include <grp.h>
@@ -37,23 +37,21 @@
 #include <unistd.h>
 #include <sys/mman.h>
 #include <assert.h>
-#include <config.h>
 #include <urcu/compiler.h>
 #include <ulimit.h>
 
 #include <common/defaults.h>
 #include <common/common.h>
-#include <common/consumer.h>
-#include <common/consumer-timer.h>
+#include <common/consumer/consumer.h>
+#include <common/consumer/consumer-timer.h>
 #include <common/compat/poll.h>
+#include <common/compat/getenv.h>
 #include <common/sessiond-comm/sessiond-comm.h>
 #include <common/utils.h>
 
 #include "lttng-consumerd.h"
 #include "health-consumerd.h"
 
-/* TODO : support UST (all direct kernel-ctl accesses). */
-
 /* threads (channel handling, poll, metadata, sessiond) */
 
 static pthread_t channel_thread, data_thread, metadata_thread,
@@ -101,15 +99,9 @@ static void sighandler(int sig)
 		return;
 	}
 
-	/*
-	 * Ignore SIGPIPE because it should not stop the consumer whenever a
-	 * SIGPIPE is catched through a FD operation.
-	 */
-	if (sig == SIGPIPE) {
-		return;
+	if (ctx) {
+		lttng_consumer_should_exit(ctx);
 	}
-
-	lttng_consumer_should_exit(ctx);
 }
 
 /*
@@ -123,25 +115,27 @@ static int set_signal_handler(void)
 	sigset_t sigset;
 
 	if ((ret = sigemptyset(&sigset)) < 0) {
-		perror("sigemptyset");
+		PERROR("sigemptyset");
 		return ret;
 	}
 
-	sa.sa_handler = sighandler;
 	sa.sa_mask = sigset;
 	sa.sa_flags = 0;
+
+	sa.sa_handler = sighandler;
 	if ((ret = sigaction(SIGTERM, &sa, NULL)) < 0) {
-		perror("sigaction");
+		PERROR("sigaction");
 		return ret;
 	}
 
 	if ((ret = sigaction(SIGINT, &sa, NULL)) < 0) {
-		perror("sigaction");
+		PERROR("sigaction");
 		return ret;
 	}
 
+	sa.sa_handler = SIG_IGN;
 	if ((ret = sigaction(SIGPIPE, &sa, NULL)) < 0) {
-		perror("sigaction");
+		PERROR("sigaction");
 		return ret;
 	}
 
@@ -185,9 +179,9 @@ static void usage(FILE *fp)
 /*
  * daemon argument parsing
  */
-static void parse_args(int argc, char **argv)
+static int parse_args(int argc, char **argv)
 {
-	int c;
+	int c, ret = 0;
 
 	static struct option long_options[] = {
 		{ "consumerd-cmd-sock", 1, 0, 'c' },
@@ -207,29 +201,48 @@ static void parse_args(int argc, char **argv)
 
 	while (1) {
 		int option_index = 0;
-		c = getopt_long(argc, argv, "dhqvVku" "c:e:g:", long_options, &option_index);
+		c = getopt_long(argc, argv, "dhqvVku" "c:e:g:",
+				long_options, &option_index);
 		if (c == -1) {
 			break;
 		}
 
 		switch (c) {
 		case 0:
-			fprintf(stderr, "option %s", long_options[option_index].name);
+			fprintf(stderr, "option %s",
+				long_options[option_index].name);
 			if (optarg) {
 				fprintf(stderr, " with arg %s\n", optarg);
+				ret = -1;
+				goto end;
 			}
 			break;
 		case 'c':
-			snprintf(command_sock_path, PATH_MAX, "%s", optarg);
+			if (lttng_is_setuid_setgid()) {
+				WARN("Getting '%s' argument from setuid/setgid binary refused for security reasons.",
+					"-c, --consumerd-cmd-sock");
+			} else {
+				snprintf(command_sock_path, PATH_MAX, "%s", optarg);
+			}
 			break;
 		case 'e':
-			snprintf(error_sock_path, PATH_MAX, "%s", optarg);
+			if (lttng_is_setuid_setgid()) {
+				WARN("Getting '%s' argument from setuid/setgid binary refused for security reasons.",
+					"-e, --consumerd-err-sock");
+			} else {
+				snprintf(error_sock_path, PATH_MAX, "%s", optarg);
+			}
 			break;
 		case 'd':
 			opt_daemon = 1;
 			break;
 		case 'g':
-			tracing_group_name = optarg;
+			if (lttng_is_setuid_setgid()) {
+				WARN("Getting '%s' argument from setuid/setgid binary refused for security reasons.",
+					"-g, --group");
+			} else {
+				tracing_group_name = optarg;
+			}
 			break;
 		case 'h':
 			usage(stdout);
@@ -259,9 +272,12 @@ static void parse_args(int argc, char **argv)
 #endif
 		default:
 			usage(stderr);
-			exit(EXIT_FAILURE);
+			ret = -1;
+			goto end;
 		}
 	}
+end:
+	return ret;
 }
 
 /*
@@ -288,12 +304,23 @@ static void set_ulimit(void)
  */
 int main(int argc, char **argv)
 {
-	int ret = 0;
+	int ret = 0, retval = 0;
 	void *status;
+	struct lttng_consumer_local_data *tmp_ctx;
+
+	rcu_register_thread();
+
+	if (set_signal_handler()) {
+		retval = -1;
+		goto exit_set_signal_handler;
+	}
 
 	/* Parse arguments */
 	progname = argv[0];
-	parse_args(argc, argv);
+	if (parse_args(argc, argv)) {
+		retval = -1;
+		goto exit_options;
+	}
 
 	/* Daemonize */
 	if (opt_daemon) {
@@ -307,7 +334,8 @@ int main(int argc, char **argv)
 		ret = daemon(0, 0);
 		if (ret < 0) {
 			PERROR("daemon");
-			goto error;
+			retval = -1;
+			goto exit_options;
 		}
 		/*
 		 * We are in the child. Make sure all other file
@@ -319,36 +347,62 @@ int main(int argc, char **argv)
 		}
 	}
 
-	/* Set up max poll set size */
-	lttng_poll_set_max_size();
+	/*
+	 * Starting from here, we can create threads. This needs to be after
+	 * lttng_daemonize due to RCU.
+	 */
+
+	health_consumerd = health_app_create(NR_HEALTH_CONSUMERD_TYPES);
+	if (!health_consumerd) {
+		retval = -1;
+		goto exit_health_consumerd_cleanup;
+	}
 
 	if (*command_sock_path == '\0') {
 		switch (opt_type) {
 		case LTTNG_CONSUMER_KERNEL:
-			snprintf(command_sock_path, PATH_MAX, DEFAULT_KCONSUMERD_CMD_SOCK_PATH,
+			ret = snprintf(command_sock_path, PATH_MAX,
+					DEFAULT_KCONSUMERD_CMD_SOCK_PATH,
 					DEFAULT_LTTNG_RUNDIR);
+			if (ret < 0) {
+				retval = -1;
+				goto exit_init_data;
+			}
 			break;
 		case LTTNG_CONSUMER64_UST:
-			snprintf(command_sock_path, PATH_MAX,
-					DEFAULT_USTCONSUMERD64_CMD_SOCK_PATH, DEFAULT_LTTNG_RUNDIR);
+			ret = snprintf(command_sock_path, PATH_MAX,
+					DEFAULT_USTCONSUMERD64_CMD_SOCK_PATH,
+					DEFAULT_LTTNG_RUNDIR);
+			if (ret < 0) {
+				retval = -1;
+				goto exit_init_data;
+			}
 			break;
 		case LTTNG_CONSUMER32_UST:
-			snprintf(command_sock_path, PATH_MAX,
-					DEFAULT_USTCONSUMERD32_CMD_SOCK_PATH, DEFAULT_LTTNG_RUNDIR);
+			ret = snprintf(command_sock_path, PATH_MAX,
+					DEFAULT_USTCONSUMERD32_CMD_SOCK_PATH,
+					DEFAULT_LTTNG_RUNDIR);
+			if (ret < 0) {
+				retval = -1;
+				goto exit_init_data;
+			}
 			break;
 		default:
-			WARN("Unknown consumerd type");
-			goto error;
+			ERR("Unknown consumerd type");
+			retval = -1;
+			goto exit_init_data;
 		}
 	}
 
 	/* Init */
-	if (lttng_consumer_init() < 0) {
-		goto error;
+	if (lttng_consumer_init()) {
+		retval = -1;
+		goto exit_init_data;
 	}
 
-	/* Init socket timeouts */
+	/* Initialize communication library */
 	lttcomm_init();
+	/* Initialize TCP timeout values */
 	lttcomm_inet_init();
 
 	if (!getuid()) {
@@ -356,47 +410,62 @@ int main(int argc, char **argv)
 		set_ulimit();
 	}
 
-	health_consumerd = health_app_create(NR_HEALTH_CONSUMERD_TYPES);
-	if (!health_consumerd) {
-		goto error;
+	if (run_as_create_worker(argv[0]) < 0) {
+		goto exit_init_data;
 	}
 
 	/* create the consumer instance with and assign the callbacks */
 	ctx = lttng_consumer_create(opt_type, lttng_consumer_read_subbuffer,
 		NULL, lttng_consumer_on_recv_stream, NULL);
-	if (ctx == NULL) {
-		goto error;
+	if (!ctx) {
+		retval = -1;
+		goto exit_init_data;
 	}
 
 	lttng_consumer_set_command_sock_path(ctx, command_sock_path);
 	if (*error_sock_path == '\0') {
 		switch (opt_type) {
 		case LTTNG_CONSUMER_KERNEL:
-			snprintf(error_sock_path, PATH_MAX, DEFAULT_KCONSUMERD_ERR_SOCK_PATH,
+			ret = snprintf(error_sock_path, PATH_MAX,
+					DEFAULT_KCONSUMERD_ERR_SOCK_PATH,
 					DEFAULT_LTTNG_RUNDIR);
+			if (ret < 0) {
+				retval = -1;
+				goto exit_init_data;
+			}
 			break;
 		case LTTNG_CONSUMER64_UST:
-			snprintf(error_sock_path, PATH_MAX,
-					DEFAULT_USTCONSUMERD64_ERR_SOCK_PATH, DEFAULT_LTTNG_RUNDIR);
+			ret = snprintf(error_sock_path, PATH_MAX,
+					DEFAULT_USTCONSUMERD64_ERR_SOCK_PATH,
+					DEFAULT_LTTNG_RUNDIR);
+			if (ret < 0) {
+				retval = -1;
+				goto exit_init_data;
+			}
 			break;
 		case LTTNG_CONSUMER32_UST:
-			snprintf(error_sock_path, PATH_MAX,
-					DEFAULT_USTCONSUMERD32_ERR_SOCK_PATH, DEFAULT_LTTNG_RUNDIR);
+			ret = snprintf(error_sock_path, PATH_MAX,
+					DEFAULT_USTCONSUMERD32_ERR_SOCK_PATH,
+					DEFAULT_LTTNG_RUNDIR);
+			if (ret < 0) {
+				retval = -1;
+				goto exit_init_data;
+			}
 			break;
 		default:
-			WARN("Unknown consumerd type");
-			goto error;
+			ERR("Unknown consumerd type");
+			retval = -1;
+			goto exit_init_data;
 		}
 	}
 
-	if (set_signal_handler() < 0) {
-		goto error;
-	}
-
 	/* Connect to the socket created by lttng-sessiond to report errors */
 	DBG("Connecting to error socket %s", error_sock_path);
 	ret = lttcomm_connect_unix_sock(error_sock_path);
-	/* not a fatal error, but all communication with lttng-sessiond will fail */
+	/*
+	 * Not a fatal error, but all communication with lttng-sessiond will
+	 * fail.
+	 */
 	if (ret < 0) {
 		WARN("Cannot connect to error socket (is lttng-sessiond started?)");
 	}
@@ -406,21 +475,26 @@ int main(int argc, char **argv)
 	 * Block RT signals used for UST periodical metadata flush and the live
 	 * timer in main, and create a dedicated thread to handle these signals.
 	 */
-	consumer_signal_init();
+	if (consumer_signal_init()) {
+		retval = -1;
+		goto exit_init_data;
+	}
 
 	ctx->type = opt_type;
 
-	ret = utils_create_pipe(health_quit_pipe);
-	if (ret < 0) {
-		goto error_health_pipe;
+	if (utils_create_pipe(health_quit_pipe)) {
+		retval = -1;
+		goto exit_health_pipe;
 	}
 
 	/* Create thread to manage the client socket */
-	ret = pthread_create(&health_thread, NULL,
+	ret = pthread_create(&health_thread, default_pthread_attr(),
 			thread_manage_health, (void *) NULL);
-	if (ret != 0) {
+	if (ret) {
+		errno = ret;
 		PERROR("pthread_create health");
-		goto health_error;
+		retval = -1;
+		goto exit_health_thread;
 	}
 
 	/*
@@ -433,107 +507,149 @@ int main(int argc, char **argv)
 	cmm_smp_mb();	/* Read ready before following operations */
 
 	/* Create thread to manage channels */
-	ret = pthread_create(&channel_thread, NULL, consumer_thread_channel_poll,
+	ret = pthread_create(&channel_thread, default_pthread_attr(),
+			consumer_thread_channel_poll,
 			(void *) ctx);
-	if (ret != 0) {
-		perror("pthread_create");
-		goto channel_error;
+	if (ret) {
+		errno = ret;
+		PERROR("pthread_create");
+		retval = -1;
+		goto exit_channel_thread;
 	}
 
 	/* Create thread to manage the polling/writing of trace metadata */
-	ret = pthread_create(&metadata_thread, NULL, consumer_thread_metadata_poll,
+	ret = pthread_create(&metadata_thread, default_pthread_attr(),
+			consumer_thread_metadata_poll,
 			(void *) ctx);
-	if (ret != 0) {
-		perror("pthread_create");
-		goto metadata_error;
+	if (ret) {
+		errno = ret;
+		PERROR("pthread_create");
+		retval = -1;
+		goto exit_metadata_thread;
 	}
 
 	/* Create thread to manage the polling/writing of trace data */
-	ret = pthread_create(&data_thread, NULL, consumer_thread_data_poll,
-			(void *) ctx);
-	if (ret != 0) {
-		perror("pthread_create");
-		goto data_error;
+	ret = pthread_create(&data_thread, default_pthread_attr(),
+			consumer_thread_data_poll, (void *) ctx);
+	if (ret) {
+		errno = ret;
+		PERROR("pthread_create");
+		retval = -1;
+		goto exit_data_thread;
 	}
 
 	/* Create the thread to manage the receive of fd */
-	ret = pthread_create(&sessiond_thread, NULL, consumer_thread_sessiond_poll,
+	ret = pthread_create(&sessiond_thread, default_pthread_attr(),
+			consumer_thread_sessiond_poll,
 			(void *) ctx);
-	if (ret != 0) {
-		perror("pthread_create");
-		goto sessiond_error;
+	if (ret) {
+		errno = ret;
+		PERROR("pthread_create");
+		retval = -1;
+		goto exit_sessiond_thread;
 	}
 
 	/*
 	 * Create the thread to manage the UST metadata periodic timer and
 	 * live timer.
 	 */
-	ret = pthread_create(&metadata_timer_thread, NULL,
+	ret = pthread_create(&metadata_timer_thread, default_pthread_attr(),
 			consumer_timer_thread, (void *) ctx);
-	if (ret != 0) {
-		perror("pthread_create");
-		goto metadata_timer_error;
+	if (ret) {
+		errno = ret;
+		PERROR("pthread_create");
+		retval = -1;
+		goto exit_metadata_timer_thread;
 	}
 
 	ret = pthread_detach(metadata_timer_thread);
 	if (ret) {
 		errno = ret;
-		perror("pthread_detach");
+		PERROR("pthread_detach");
+		retval = -1;
+		goto exit_metadata_timer_detach;
 	}
 
-metadata_timer_error:
+	/*
+	 * This is where we start awaiting program completion (e.g. through
+	 * signal that asks threads to teardown.
+	 */
+
+exit_metadata_timer_detach:
+exit_metadata_timer_thread:
 	ret = pthread_join(sessiond_thread, &status);
-	if (ret != 0) {
-		perror("pthread_join");
-		goto error;
+	if (ret) {
+		errno = ret;
+		PERROR("pthread_join sessiond_thread");
+		retval = -1;
 	}
+exit_sessiond_thread:
 
-sessiond_error:
 	ret = pthread_join(data_thread, &status);
-	if (ret != 0) {
-		perror("pthread_join");
-		goto error;
+	if (ret) {
+		errno = ret;
+		PERROR("pthread_join data_thread");
+		retval = -1;
 	}
+exit_data_thread:
 
-data_error:
 	ret = pthread_join(metadata_thread, &status);
-	if (ret != 0) {
-		perror("pthread_join");
-		goto error;
+	if (ret) {
+		errno = ret;
+		PERROR("pthread_join metadata_thread");
+		retval = -1;
 	}
+exit_metadata_thread:
 
-metadata_error:
 	ret = pthread_join(channel_thread, &status);
-	if (ret != 0) {
-		perror("pthread_join");
-		goto error;
+	if (ret) {
+		errno = ret;
+		PERROR("pthread_join channel_thread");
+		retval = -1;
 	}
+exit_channel_thread:
 
-channel_error:
 	ret = pthread_join(health_thread, &status);
-	if (ret != 0) {
-		PERROR("pthread_join health thread");
-		goto error;	/* join error, exit without cleanup */
+	if (ret) {
+		errno = ret;
+		PERROR("pthread_join health_thread");
+		retval = -1;
 	}
+exit_health_thread:
 
-health_error:
 	utils_close_pipe(health_quit_pipe);
+exit_health_pipe:
 
-error_health_pipe:
-	if (!ret) {
-		ret = EXIT_SUCCESS;
-		goto end;
-	}
-
-error:
-	ret = EXIT_FAILURE;
-
-end:
-	lttng_consumer_destroy(ctx);
+exit_init_data:
+	tmp_ctx = ctx;
+	ctx = NULL;
+	cmm_barrier();	/* Clear ctx for signal handler. */
+	/*
+	 * Wait for all pending call_rcu work to complete before tearing
+	 * down data structures. call_rcu worker may be trying to
+	 * perform lookups in those structures.
+	 */
+	rcu_barrier();
+	lttng_consumer_destroy(tmp_ctx);
 	lttng_consumer_cleanup();
+
 	if (health_consumerd) {
 		health_app_destroy(health_consumerd);
 	}
+	/* Ensure all prior call_rcu are done. */
+	rcu_barrier();
 
-	return ret;
+	run_as_destroy_worker();
+
+exit_health_consumerd_cleanup:
+exit_options:
+exit_set_signal_handler:
+
+	rcu_unregister_thread();
+
+	if (!retval) {
+		exit(EXIT_SUCCESS);
+	} else {
+		exit(EXIT_FAILURE);
+	}
 }
