@@ -15,78 +15,57 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
-#define _GNU_SOURCE
-#include <errno.h>
+#define _LGPL_SOURCE
 #include <fcntl.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
 #include <unistd.h>
+#include <inttypes.h>
 
 #include <common/common.h>
 #include <common/kernel-ctl/kernel-ctl.h>
+#include <common/kernel-ctl/kernel-ioctl.h>
 #include <common/sessiond-comm/sessiond-comm.h>
 
+#include "consumer.h"
 #include "kernel.h"
+#include "kernel-consumer.h"
 #include "kern-modules.h"
+#include "utils.h"
 
 /*
  * Add context on a kernel channel.
  */
 int kernel_add_channel_context(struct ltt_kernel_channel *chan,
-		struct lttng_kernel_context *ctx)
+		struct ltt_kernel_context *ctx)
 {
 	int ret;
 
+	assert(chan);
+	assert(ctx);
+
 	DBG("Adding context to channel %s", chan->channel->name);
-	ret = kernctl_add_context(chan->fd, ctx);
+	ret = kernctl_add_context(chan->fd, &ctx->ctx);
 	if (ret < 0) {
-		if (errno != EEXIST) {
-			PERROR("add context ioctl");
-		} else {
+		switch (-ret) {
+		case ENOSYS:
+			/* Exists but not available for this kernel */
+			ret = LTTNG_ERR_KERN_CONTEXT_UNAVAILABLE;
+			goto error;
+		case EEXIST:
 			/* If EEXIST, we just ignore the error */
 			ret = 0;
+			goto end;
+		default:
+			PERROR("add context ioctl");
+			ret = LTTNG_ERR_KERN_CONTEXT_FAIL;
+			goto error;
 		}
-		goto error;
 	}
 
-	chan->ctx = zmalloc(sizeof(struct lttng_kernel_context));
-	if (chan->ctx == NULL) {
-		PERROR("zmalloc event context");
-		goto error;
-	}
-
-	memcpy(chan->ctx, ctx, sizeof(struct lttng_kernel_context));
-
-	return 0;
-
-error:
-	return ret;
-}
-
-/*
- * Add context on a kernel event.
- */
-int kernel_add_event_context(struct ltt_kernel_event *event,
-		struct lttng_kernel_context *ctx)
-{
-	int ret;
-
-	DBG("Adding context to event %s", event->event->name);
-	ret = kernctl_add_context(event->fd, ctx);
-	if (ret < 0) {
-		PERROR("add context ioctl");
-		goto error;
-	}
-
-	event->ctx = zmalloc(sizeof(struct lttng_kernel_context));
-	if (event->ctx == NULL) {
-		PERROR("zmalloc event context");
-		goto error;
-	}
-
-	memcpy(event->ctx, ctx, sizeof(struct lttng_kernel_context));
-
+end:
+	cds_list_add_tail(&ctx->list, &chan->ctx_list);
 	return 0;
 
 error:
@@ -102,8 +81,10 @@ int kernel_create_session(struct ltt_session *session, int tracer_fd)
 	int ret;
 	struct ltt_kernel_session *lks;
 
+	assert(session);
+
 	/* Allocate data structure */
-	lks = trace_kernel_create_session(session->path);
+	lks = trace_kernel_create_session();
 	if (lks == NULL) {
 		ret = -1;
 		goto error;
@@ -123,6 +104,7 @@ int kernel_create_session(struct ltt_session *session, int tracer_fd)
 		PERROR("fcntl session fd");
 	}
 
+	lks->id = session->id;
 	lks->consumer_fds_sent = 0;
 	session->kernel_session = lks;
 
@@ -131,6 +113,9 @@ int kernel_create_session(struct ltt_session *session, int tracer_fd)
 	return 0;
 
 error:
+	if (lks) {
+		trace_kernel_destroy_session(lks);
+	}
 	return ret;
 }
 
@@ -139,17 +124,26 @@ error:
  * kernel session.
  */
 int kernel_create_channel(struct ltt_kernel_session *session,
-		struct lttng_channel *chan, char *path)
+		struct lttng_channel *chan)
 {
 	int ret;
 	struct ltt_kernel_channel *lkc;
 
+	assert(session);
+	assert(chan);
+
 	/* Allocate kernel channel */
-	lkc = trace_kernel_create_channel(chan, path);
+	lkc = trace_kernel_create_channel(chan);
 	if (lkc == NULL) {
 		goto error;
 	}
 
+	DBG3("Kernel create channel %s with attr: %d, %" PRIu64 ", %" PRIu64 ", %u, %u, %d, %d",
+			chan->name, lkc->channel->attr.overwrite,
+			lkc->channel->attr.subbuf_size, lkc->channel->attr.num_subbuf,
+			lkc->channel->attr.switch_timer_interval, lkc->channel->attr.read_timer_interval,
+			lkc->channel->attr.live_timer_interval, lkc->channel->attr.output);
+
 	/* Kernel tracer channel creation */
 	ret = kernctl_create_channel(session->fd, &lkc->channel->attr);
 	if (ret < 0) {
@@ -168,27 +162,39 @@ int kernel_create_channel(struct ltt_kernel_session *session,
 	/* Add channel to session */
 	cds_list_add(&lkc->list, &session->channel_list.head);
 	session->channel_count++;
+	lkc->session = session;
 
-	DBG("Kernel channel %s created (fd: %d and path: %s)",
-			lkc->channel->name, lkc->fd, lkc->pathname);
+	DBG("Kernel channel %s created (fd: %d)", lkc->channel->name, lkc->fd);
 
 	return 0;
 
 error:
+	if (lkc) {
+		free(lkc->channel);
+		free(lkc);
+	}
 	return -1;
 }
 
 /*
  * Create a kernel event, enable it to the kernel tracer and add it to the
  * channel event list of the kernel session.
+ * We own filter_expression and filter.
  */
 int kernel_create_event(struct lttng_event *ev,
-		struct ltt_kernel_channel *channel)
+		struct ltt_kernel_channel *channel,
+		char *filter_expression,
+		struct lttng_filter_bytecode *filter)
 {
 	int ret;
 	struct ltt_kernel_event *event;
 
-	event = trace_kernel_create_event(ev);
+	assert(ev);
+	assert(channel);
+
+	/* We pass ownership of filter_expression and filter */
+	event = trace_kernel_create_event(ev, filter_expression,
+			filter);
 	if (event == NULL) {
 		ret = -1;
 		goto error;
@@ -196,32 +202,22 @@ int kernel_create_event(struct lttng_event *ev,
 
 	ret = kernctl_create_event(channel->fd, event->event);
 	if (ret < 0) {
-		switch (errno) {
+		switch (-ret) {
 		case EEXIST:
 			break;
 		case ENOSYS:
 			WARN("Event type not implemented");
 			break;
+		case ENOENT:
+			WARN("Event %s not found!", ev->name);
+			break;
 		default:
 			PERROR("create event ioctl");
 		}
-		ret = -errno;
 		goto free_event;
 	}
 
-	/*
-	 * LTTNG_KERNEL_SYSCALL event creation will return 0 on success.
-	 */
-	if (ret == 0 && event->event->instrumentation == LTTNG_KERNEL_SYSCALL) {
-		DBG2("Kernel event syscall creation success");
-		/*
-		 * We use fd == -1 to ensure that we never trigger a close of fd
-		 * 0.
-		 */
-		event->fd = -1;
-		goto add_list;
-	}
-
+	event->type = ev->type;
 	event->fd = ret;
 	/* Prevent fd duplication after execlp() */
 	ret = fcntl(event->fd, F_SETFD, FD_CLOEXEC);
@@ -229,7 +225,26 @@ int kernel_create_event(struct lttng_event *ev,
 		PERROR("fcntl session fd");
 	}
 
-add_list:
+	if (filter) {
+		ret = kernctl_filter(event->fd, filter);
+		if (ret) {
+			goto filter_error;
+		}
+	}
+
+	ret = kernctl_enable(event->fd);
+	if (ret < 0) {
+		switch (-ret) {
+		case EEXIST:
+			ret = LTTNG_ERR_KERN_EVENT_EXIST;
+			break;
+		default:
+			PERROR("enable kernel event");
+			break;
+		}
+		goto enable_error;
+	}
+
 	/* Add event to event list */
 	cds_list_add(&event->list, &channel->events_list.head);
 	channel->event_count++;
@@ -238,6 +253,16 @@ add_list:
 
 	return 0;
 
+enable_error:
+filter_error:
+	{
+		int closeret;
+
+		closeret = close(event->fd);
+		if (closeret) {
+			PERROR("close event fd");
+		}
+	}
 free_event:
 	free(event);
 error:
@@ -251,10 +276,11 @@ int kernel_disable_channel(struct ltt_kernel_channel *chan)
 {
 	int ret;
 
+	assert(chan);
+
 	ret = kernctl_disable(chan->fd);
 	if (ret < 0) {
 		PERROR("disable chan ioctl");
-		ret = errno;
 		goto error;
 	}
 
@@ -274,8 +300,10 @@ int kernel_enable_channel(struct ltt_kernel_channel *chan)
 {
 	int ret;
 
+	assert(chan);
+
 	ret = kernctl_enable(chan->fd);
-	if (ret < 0 && errno != EEXIST) {
+	if (ret < 0 && ret != -EEXIST) {
 		PERROR("Enable kernel chan");
 		goto error;
 	}
@@ -296,11 +324,13 @@ int kernel_enable_event(struct ltt_kernel_event *event)
 {
 	int ret;
 
+	assert(event);
+
 	ret = kernctl_enable(event->fd);
 	if (ret < 0) {
-		switch (errno) {
+		switch (-ret) {
 		case EEXIST:
-			ret = LTTCOMM_KERN_EVENT_EXIST;
+			ret = LTTNG_ERR_KERN_EVENT_EXIST;
 			break;
 		default:
 			PERROR("enable kernel event");
@@ -325,11 +355,13 @@ int kernel_disable_event(struct ltt_kernel_event *event)
 {
 	int ret;
 
+	assert(event);
+
 	ret = kernctl_disable(event->fd);
 	if (ret < 0) {
-		switch (errno) {
+		switch (-ret) {
 		case EEXIST:
-			ret = LTTCOMM_KERN_EVENT_EXIST;
+			ret = LTTNG_ERR_KERN_EVENT_EXIST;
 			break;
 		default:
 			PERROR("disable kernel event");
@@ -347,17 +379,135 @@ error:
 	return ret;
 }
 
+
+int kernel_track_pid(struct ltt_kernel_session *session, int pid)
+{
+	int ret;
+
+	DBG("Kernel track PID %d for session id %" PRIu64 ".",
+			pid, session->id);
+	ret = kernctl_track_pid(session->fd, pid);
+	if (!ret) {
+		return LTTNG_OK;
+	}
+	switch (-ret) {
+	case EINVAL:
+		return LTTNG_ERR_INVALID;
+	case ENOMEM:
+		return LTTNG_ERR_NOMEM;
+	case EEXIST:
+		return LTTNG_ERR_PID_TRACKED;
+	default:
+		return LTTNG_ERR_UNK;
+	}
+}
+
+int kernel_untrack_pid(struct ltt_kernel_session *session, int pid)
+{
+	int ret;
+
+	DBG("Kernel untrack PID %d for session id %" PRIu64 ".",
+			pid, session->id);
+	ret = kernctl_untrack_pid(session->fd, pid);
+	if (!ret) {
+		return LTTNG_OK;
+	}
+	switch (-ret) {
+	case EINVAL:
+		return LTTNG_ERR_INVALID;
+	case ENOMEM:
+		return LTTNG_ERR_NOMEM;
+	case ENOENT:
+		return LTTNG_ERR_PID_NOT_TRACKED;
+	default:
+		return LTTNG_ERR_UNK;
+	}
+}
+
+ssize_t kernel_list_tracker_pids(struct ltt_kernel_session *session,
+		int **_pids)
+{
+	int fd, ret;
+	int pid;
+	ssize_t nbmem, count = 0;
+	FILE *fp;
+	int *pids;
+
+	fd = kernctl_list_tracker_pids(session->fd);
+	if (fd < 0) {
+		PERROR("kernel tracker pids list");
+		goto error;
+	}
+
+	fp = fdopen(fd, "r");
+	if (fp == NULL) {
+		PERROR("kernel tracker pids list fdopen");
+		goto error_fp;
+	}
+
+	nbmem = KERNEL_TRACKER_PIDS_INIT_LIST_SIZE;
+	pids = zmalloc(sizeof(*pids) * nbmem);
+	if (pids == NULL) {
+		PERROR("alloc list pids");
+		count = -ENOMEM;
+		goto end;
+	}
+
+	while (fscanf(fp, "process { pid = %u; };\n", &pid) == 1) {
+		if (count >= nbmem) {
+			int *new_pids;
+			size_t new_nbmem;
+
+			new_nbmem = nbmem << 1;
+			DBG("Reallocating pids list from %zu to %zu entries",
+					nbmem, new_nbmem);
+			new_pids = realloc(pids, new_nbmem * sizeof(*new_pids));
+			if (new_pids == NULL) {
+				PERROR("realloc list events");
+				free(pids);
+				count = -ENOMEM;
+				goto end;
+			}
+			/* Zero the new memory */
+			memset(new_pids + nbmem, 0,
+				(new_nbmem - nbmem) * sizeof(*new_pids));
+			nbmem = new_nbmem;
+			pids = new_pids;
+		}
+		pids[count++] = pid;
+	}
+
+	*_pids = pids;
+	DBG("Kernel list tracker pids done (%zd pids)", count);
+end:
+	ret = fclose(fp);	/* closes both fp and fd */
+	if (ret) {
+		PERROR("fclose");
+	}
+	return count;
+
+error_fp:
+	ret = close(fd);
+	if (ret) {
+		PERROR("close");
+	}
+error:
+	return -1;
+}
+
 /*
  * Create kernel metadata, open from the kernel tracer and add it to the
  * kernel session.
  */
-int kernel_open_metadata(struct ltt_kernel_session *session, char *path)
+int kernel_open_metadata(struct ltt_kernel_session *session)
 {
 	int ret;
-	struct ltt_kernel_metadata *lkm;
+	struct ltt_kernel_metadata *lkm = NULL;
+
+	assert(session);
 
 	/* Allocate kernel metadata */
-	lkm = trace_kernel_create_metadata(path);
+	lkm = trace_kernel_create_metadata();
 	if (lkm == NULL) {
 		goto error;
 	}
@@ -365,7 +515,7 @@ int kernel_open_metadata(struct ltt_kernel_session *session, char *path)
 	/* Kernel tracer metadata creation */
 	ret = kernctl_open_metadata(session->fd, &lkm->conf->attr);
 	if (ret < 0) {
-		goto error;
+		goto error_open;
 	}
 
 	lkm->fd = ret;
@@ -377,10 +527,12 @@ int kernel_open_metadata(struct ltt_kernel_session *session, char *path)
 
 	session->metadata = lkm;
 
-	DBG("Kernel metadata opened (fd: %d and path: %s)", lkm->fd, lkm->pathname);
+	DBG("Kernel metadata opened (fd: %d)", lkm->fd);
 
 	return 0;
 
+error_open:
+	trace_kernel_destroy_metadata(lkm);
 error:
 	return -1;
 }
@@ -392,6 +544,8 @@ int kernel_start_session(struct ltt_kernel_session *session)
 {
 	int ret;
 
+	assert(session);
+
 	ret = kernctl_start_session(session->fd);
 	if (ret < 0) {
 		PERROR("ioctl start session");
@@ -422,23 +576,6 @@ void kernel_wait_quiescent(int fd)
 	}
 }
 
-/*
- * Kernel calibrate
- */
-int kernel_calibrate(int fd, struct lttng_kernel_calibrate *calibrate)
-{
-	int ret;
-
-	ret = kernctl_calibrate(fd, calibrate);
-	if (ret < 0) {
-		PERROR("calibrate ioctl");
-		return -1;
-	}
-
-	return 0;
-}
-
-
 /*
  *  Force flush buffer of metadata.
  */
@@ -446,9 +583,11 @@ int kernel_metadata_flush_buffer(int fd)
 {
 	int ret;
 
+	DBG("Kernel flushing metadata buffer on fd %d", fd);
+
 	ret = kernctl_buffer_flush(fd);
 	if (ret < 0) {
-		ERR("Fail to flush metadata buffers %d (ret: %d", fd, ret);
+		ERR("Fail to flush metadata buffers %d (ret: %d)", fd, ret);
 	}
 
 	return 0;
@@ -462,6 +601,8 @@ int kernel_flush_buffer(struct ltt_kernel_channel *channel)
 	int ret;
 	struct ltt_kernel_stream *stream;
 
+	assert(channel);
+
 	DBG("Flush buffer for channel %s", channel->channel->name);
 
 	cds_list_for_each_entry(stream, &channel->stream_list.head, list) {
@@ -484,6 +625,8 @@ int kernel_stop_session(struct ltt_kernel_session *session)
 {
 	int ret;
 
+	assert(session);
+
 	ret = kernctl_stop_session(session->fd);
 	if (ret < 0) {
 		goto error;
@@ -501,6 +644,10 @@ error:
  * Open stream of channel, register it to the kernel tracer and add it
  * to the stream list of the channel.
  *
+ * Note: given that the streams may appear in random order wrt CPU
+ * number (e.g. cpu hotplug), the index value of the stream number in
+ * the stream name is not necessarily linked to the CPU number.
+ *
  * Return the number of created stream. Else, a negative value.
  */
 int kernel_open_channel_stream(struct ltt_kernel_channel *channel)
@@ -508,8 +655,11 @@ int kernel_open_channel_stream(struct ltt_kernel_channel *channel)
 	int ret;
 	struct ltt_kernel_stream *lks;
 
+	assert(channel);
+
 	while ((ret = kernctl_create_stream(channel->fd)) >= 0) {
-		lks = trace_kernel_create_stream();
+		lks = trace_kernel_create_stream(channel->channel->name,
+				channel->stream_count);
 		if (lks == NULL) {
 			ret = close(ret);
 			if (ret) {
@@ -525,19 +675,15 @@ int kernel_open_channel_stream(struct ltt_kernel_channel *channel)
 			PERROR("fcntl session fd");
 		}
 
-		ret = asprintf(&lks->pathname, "%s/%s_%d",
-				channel->pathname, channel->channel->name, channel->stream_count);
-		if (ret < 0) {
-			PERROR("asprintf kernel create stream");
-			goto error;
-		}
+		lks->tracefile_size = channel->channel->attr.tracefile_size;
+		lks->tracefile_count = channel->channel->attr.tracefile_count;
 
-		/* Add stream to channe stream list */
+		/* Add stream to channel stream list */
 		cds_list_add(&lks->list, &channel->stream_list.head);
 		channel->stream_count++;
 
-		DBG("Kernel stream %d created (fd: %d, state: %d, path: %s)",
-				channel->stream_count, lks->fd, lks->state, lks->pathname);
+		DBG("Kernel stream %s created (fd: %d, state: %d)", lks->name, lks->fd,
+				lks->state);
 	}
 
 	return channel->stream_count;
@@ -553,6 +699,8 @@ int kernel_open_metadata_stream(struct ltt_kernel_session *session)
 {
 	int ret;
 
+	assert(session);
+
 	ret = kernctl_create_stream(session->metadata->fd);
 	if (ret < 0) {
 		PERROR("kernel create metadata stream");
@@ -578,13 +726,14 @@ error:
  */
 ssize_t kernel_list_events(int tracer_fd, struct lttng_event **events)
 {
-	int fd, pos, ret;
+	int fd, ret;
 	char *event;
 	size_t nbmem, count = 0;
-	ssize_t size;
 	FILE *fp;
 	struct lttng_event *elist;
 
+	assert(events);
+
 	fd = kernctl_tracepoint_list(tracer_fd);
 	if (fd < 0) {
 		PERROR("kernel tracepoint list");
@@ -609,15 +758,15 @@ ssize_t kernel_list_events(int tracer_fd, struct lttng_event **events)
 		goto end;
 	}
 
-	while ((size = fscanf(fp, "event { name = %m[^;]; };%n\n", &event, &pos)) == 1) {
+	while (fscanf(fp, "event { name = %m[^;]; };\n", &event) == 1) {
 		if (count >= nbmem) {
 			struct lttng_event *new_elist;
+			size_t new_nbmem;
 
-			DBG("Reallocating event list from %zu to %zu bytes", nbmem,
-					nbmem * 2);
-			/* Double the size */
-			nbmem <<= 1;
-			new_elist = realloc(elist, nbmem * sizeof(struct lttng_event));
+			new_nbmem = nbmem << 1;
+			DBG("Reallocating event list from %zu to %zu bytes",
+					nbmem, new_nbmem);
+			new_elist = realloc(elist, new_nbmem * sizeof(struct lttng_event));
 			if (new_elist == NULL) {
 				PERROR("realloc list events");
 				free(event);
@@ -625,6 +774,10 @@ ssize_t kernel_list_events(int tracer_fd, struct lttng_event **events)
 				count = -ENOMEM;
 				goto end;
 			}
+			/* Zero the new memory */
+			memset(new_elist + nbmem, 0,
+				(new_nbmem - nbmem) * sizeof(struct lttng_event));
+			nbmem = new_nbmem;
 			elist = new_elist;
 		}
 		strncpy(elist[count].name, event, LTTNG_SYMBOL_NAME_LEN);
@@ -659,28 +812,41 @@ int kernel_validate_version(int tracer_fd)
 {
 	int ret;
 	struct lttng_kernel_tracer_version version;
+	struct lttng_kernel_tracer_abi_version abi_version;
 
 	ret = kernctl_tracer_version(tracer_fd, &version);
 	if (ret < 0) {
-		ERR("Failed at getting the lttng-modules version");
+		ERR("Failed to retrieve the lttng-modules version");
 		goto error;
 	}
 
 	/* Validate version */
-	if (version.major != KERN_MODULES_PRE_MAJOR
-		&& version.major != KERN_MODULES_MAJOR) {
+	if (version.major != VERSION_MAJOR) {
+		ERR("Kernel tracer major version (%d) is not compatible with lttng-tools major version (%d)",
+			version.major, VERSION_MAJOR);
 		goto error_version;
 	}
-
-	DBG2("Kernel tracer version validated (major version %d)", version.major);
+	ret = kernctl_tracer_abi_version(tracer_fd, &abi_version);
+	if (ret < 0) {
+		ERR("Failed to retrieve lttng-modules ABI version");
+		goto error;
+	}
+	if (abi_version.major != LTTNG_MODULES_ABI_MAJOR_VERSION) {
+		ERR("Kernel tracer ABI version (%d.%d) does not match the expected ABI major version (%d.*)",
+			abi_version.major, abi_version.minor,
+			LTTNG_MODULES_ABI_MAJOR_VERSION);
+		goto error;
+	}
+	DBG2("Kernel tracer version validated (%d.%d, ABI %d.%d)",
+			version.major, version.minor,
+			abi_version.major, abi_version.minor);
 	return 0;
 
 error_version:
-	ERR("Kernel major version %d is not compatible (supporting <= %d)",
-			version.major, KERN_MODULES_MAJOR)
 	ret = -1;
 
 error:
+	ERR("Kernel tracer version check failed; kernel tracing will not be available");
 	return ret;
 }
 
@@ -716,3 +882,209 @@ int init_kernel_workarounds(void)
 end_boot_id:
 	return 0;
 }
+
+/*
+ * Complete teardown of a kernel session.
+ */
+void kernel_destroy_session(struct ltt_kernel_session *ksess)
+{
+	if (ksess == NULL) {
+		DBG3("No kernel session when tearing down session");
+		return;
+	}
+
+	DBG("Tearing down kernel session");
+
+	/*
+	 * Destroy channels on the consumer if at least one FD has been sent and we
+	 * are in no output mode because the streams are in *no* monitor mode so we
+	 * have to send a command to clean them up or else they leaked.
+	 */
+	if (!ksess->output_traces && ksess->consumer_fds_sent) {
+		int ret;
+		struct consumer_socket *socket;
+		struct lttng_ht_iter iter;
+
+		/* For each consumer socket. */
+		rcu_read_lock();
+		cds_lfht_for_each_entry(ksess->consumer->socks->ht, &iter.iter,
+				socket, node.node) {
+			struct ltt_kernel_channel *chan;
+
+			/* For each channel, ask the consumer to destroy it. */
+			cds_list_for_each_entry(chan, &ksess->channel_list.head, list) {
+				ret = kernel_consumer_destroy_channel(socket, chan);
+				if (ret < 0) {
+					/* Consumer is probably dead. Use next socket. */
+					continue;
+				}
+			}
+		}
+		rcu_read_unlock();
+	}
+
+	/* Close any relayd session */
+	consumer_output_send_destroy_relayd(ksess->consumer);
+
+	trace_kernel_destroy_session(ksess);
+}
+
+/*
+ * Destroy a kernel channel object. It does not do anything on the tracer side.
+ */
+void kernel_destroy_channel(struct ltt_kernel_channel *kchan)
+{
+	struct ltt_kernel_session *ksess = NULL;
+
+	assert(kchan);
+	assert(kchan->channel);
+
+	DBG3("Kernel destroy channel %s", kchan->channel->name);
+
+	/* Update channel count of associated session. */
+	if (kchan->session) {
+		/* Keep pointer reference so we can update it after the destroy. */
+		ksess = kchan->session;
+	}
+
+	trace_kernel_destroy_channel(kchan);
+
+	/*
+	 * At this point the kernel channel is not visible anymore. This is safe
+	 * since in order to work on a visible kernel session, the tracing session
+	 * lock (ltt_session.lock) MUST be acquired.
+	 */
+	if (ksess) {
+		ksess->channel_count--;
+	}
+}
+
+/*
+ * Take a snapshot for a given kernel session.
+ *
+ * Return 0 on success or else return a LTTNG_ERR code.
+ */
+int kernel_snapshot_record(struct ltt_kernel_session *ksess,
+		struct snapshot_output *output, int wait,
+		uint64_t nb_packets_per_stream)
+{
+	int err, ret, saved_metadata_fd;
+	struct consumer_socket *socket;
+	struct lttng_ht_iter iter;
+	struct ltt_kernel_metadata *saved_metadata;
+
+	assert(ksess);
+	assert(ksess->consumer);
+	assert(output);
+
+	DBG("Kernel snapshot record started");
+
+	/* Save current metadata since the following calls will change it. */
+	saved_metadata = ksess->metadata;
+	saved_metadata_fd = ksess->metadata_stream_fd;
+
+	rcu_read_lock();
+
+	ret = kernel_open_metadata(ksess);
+	if (ret < 0) {
+		ret = LTTNG_ERR_KERN_META_FAIL;
+		goto error;
+	}
+
+	ret = kernel_open_metadata_stream(ksess);
+	if (ret < 0) {
+		ret = LTTNG_ERR_KERN_META_FAIL;
+		goto error_open_stream;
+	}
+
+	/* Send metadata to consumer and snapshot everything. */
+	cds_lfht_for_each_entry(ksess->consumer->socks->ht, &iter.iter,
+			socket, node.node) {
+		struct consumer_output *saved_output;
+		struct ltt_kernel_channel *chan;
+
+		/*
+		 * Temporarly switch consumer output for our snapshot output. As long
+		 * as the session lock is taken, this is safe.
+		 */
+		saved_output = ksess->consumer;
+		ksess->consumer = output->consumer;
+
+		pthread_mutex_lock(socket->lock);
+		/* This stream must not be monitored by the consumer. */
+		ret = kernel_consumer_add_metadata(socket, ksess, 0);
+		pthread_mutex_unlock(socket->lock);
+		/* Put back the saved consumer output into the session. */
+		ksess->consumer = saved_output;
+		if (ret < 0) {
+			ret = LTTNG_ERR_KERN_CONSUMER_FAIL;
+			goto error_consumer;
+		}
+
+		/* For each channel, ask the consumer to snapshot it. */
+		cds_list_for_each_entry(chan, &ksess->channel_list.head, list) {
+			pthread_mutex_lock(socket->lock);
+			ret = consumer_snapshot_channel(socket, chan->fd, output, 0,
+					ksess->uid, ksess->gid,
+					DEFAULT_KERNEL_TRACE_DIR, wait,
+					nb_packets_per_stream);
+			pthread_mutex_unlock(socket->lock);
+			if (ret < 0) {
+				ret = LTTNG_ERR_KERN_CONSUMER_FAIL;
+				(void) kernel_consumer_destroy_metadata(socket,
+						ksess->metadata);
+				goto error_consumer;
+			}
+		}
+
+		/* Snapshot metadata, */
+		pthread_mutex_lock(socket->lock);
+		ret = consumer_snapshot_channel(socket, ksess->metadata->fd, output,
+				1, ksess->uid, ksess->gid,
+				DEFAULT_KERNEL_TRACE_DIR, wait, 0);
+		pthread_mutex_unlock(socket->lock);
+		if (ret < 0) {
+			ret = LTTNG_ERR_KERN_CONSUMER_FAIL;
+			goto error_consumer;
+		}
+
+		/*
+		 * The metadata snapshot is done, ask the consumer to destroy it since
+		 * it's not monitored on the consumer side.
+		 */
+		(void) kernel_consumer_destroy_metadata(socket, ksess->metadata);
+	}
+
+	ret = LTTNG_OK;
+
+error_consumer:
+	/* Close newly opened metadata stream. It's now on the consumer side. */
+	err = close(ksess->metadata_stream_fd);
+	if (err < 0) {
+		PERROR("close snapshot kernel");
+	}
+
+error_open_stream:
+	trace_kernel_destroy_metadata(ksess->metadata);
+error:
+	/* Restore metadata state.*/
+	ksess->metadata = saved_metadata;
+	ksess->metadata_stream_fd = saved_metadata_fd;
+
+	rcu_read_unlock();
+	return ret;
+}
+
+/*
+ * Get the syscall mask array from the kernel tracer.
+ *
+ * Return 0 on success else a negative value. In both case, syscall_mask should
+ * be freed.
+ */
+int kernel_syscall_mask(int chan_fd, char **syscall_mask, uint32_t *nr_bits)
+{
+	assert(syscall_mask);
+	assert(nr_bits);
+
+	return kernctl_syscall_mask(chan_fd, syscall_mask, nr_bits);
+}
