@@ -15,7 +15,7 @@
  * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
-#define _GNU_SOURCE
+#define _LGPL_SOURCE
 #include <inttypes.h>
 #include <popt.h>
 #include <stdio.h>
@@ -24,6 +24,7 @@
 #include <assert.h>
 
 #include <common/mi-lttng.h>
+#include <lttng/constant.h>
 
 #include "../command.h"
 
@@ -31,15 +32,11 @@ static int opt_userspace;
 static int opt_kernel;
 static int opt_jul;
 static int opt_log4j;
+static int opt_python;
 static char *opt_channel;
 static int opt_domain;
 static int opt_fields;
 static int opt_syscall;
-#if 0
-/* Not implemented yet */
-static char *opt_cmd_name;
-static pid_t opt_pid;
-#endif
 
 const char *indent4 = "    ";
 const char *indent6 = "      ";
@@ -56,55 +53,20 @@ static struct mi_writer *writer;
 
 static struct poptOption long_options[] = {
 	/* longName, shortName, argInfo, argPtr, value, descrip, argDesc */
-	{"help",      'h', POPT_ARG_NONE, 0, OPT_HELP, 0, 0},
-	{"kernel",    'k', POPT_ARG_VAL, &opt_kernel, 1, 0, 0},
-	{"jul",       'j', POPT_ARG_VAL, &opt_jul, 1, 0, 0},
-	{"log4j",     'l', POPT_ARG_VAL, &opt_log4j, 1, 0, 0},
-#if 0
-	/* Not implemented yet */
-	{"userspace",      'u', POPT_ARG_STRING | POPT_ARGFLAG_OPTIONAL, &opt_cmd_name, OPT_USERSPACE, 0, 0},
-	{"pid",            'p', POPT_ARG_INT, &opt_pid, 0, 0, 0},
-#else
-	{"userspace",      'u', POPT_ARG_NONE, 0, OPT_USERSPACE, 0, 0},
-#endif
-	{"channel",   'c', POPT_ARG_STRING, &opt_channel, 0, 0, 0},
-	{"domain",    'd', POPT_ARG_VAL, &opt_domain, 1, 0, 0},
-	{"fields",    'f', POPT_ARG_VAL, &opt_fields, 1, 0, 0},
-	{"syscall",   'S', POPT_ARG_VAL, &opt_syscall, 1, 0, 0},
+	{"help",	'h', POPT_ARG_NONE, 0, OPT_HELP, 0, 0},
+	{"kernel",	'k', POPT_ARG_VAL, &opt_kernel, 1, 0, 0},
+	{"jul",	'j', POPT_ARG_VAL, &opt_jul, 1, 0, 0},
+	{"log4j",	'l', POPT_ARG_VAL, &opt_log4j, 1, 0, 0},
+	{"python",	'p', POPT_ARG_VAL, &opt_python, 1, 0, 0},
+	{"userspace",	'u', POPT_ARG_NONE, 0, OPT_USERSPACE, 0, 0},
+	{"channel",	'c', POPT_ARG_STRING, &opt_channel, 0, 0, 0},
+	{"domain",	'd', POPT_ARG_VAL, &opt_domain, 1, 0, 0},
+	{"fields",	'f', POPT_ARG_VAL, &opt_fields, 1, 0, 0},
+	{"syscall",	'S', POPT_ARG_VAL, &opt_syscall, 1, 0, 0},
 	{"list-options", 0, POPT_ARG_NONE, NULL, OPT_LIST_OPTIONS, NULL, NULL},
 	{0, 0, 0, 0, 0, 0, 0}
 };
 
-/*
- * usage
- */
-static void usage(FILE *ofp)
-{
-	fprintf(ofp, "usage: lttng list [OPTIONS] [SESSION [SESSION OPTIONS]]\n");
-	fprintf(ofp, "\n");
-	fprintf(ofp, "With no arguments, list available tracing session(s)\n");
-	fprintf(ofp, "\n");
-	fprintf(ofp, "Without a session, -k lists available kernel events\n");
-	fprintf(ofp, "Without a session, -u lists available userspace events\n");
-	fprintf(ofp, "\n");
-	fprintf(ofp, "  -h, --help              Show this help\n");
-	fprintf(ofp, "      --list-options      Simple listing of options\n");
-	fprintf(ofp, "  -k, --kernel            Select kernel domain\n");
-	fprintf(ofp, "  -u, --userspace         Select user-space domain.\n");
-	fprintf(ofp, "  -j, --jul               Apply for Java application using JUL\n");
-	fprintf(ofp, "  -l, --log4j             Apply for Java application using LOG4J\n");
-	fprintf(ofp, "  -f, --fields            List event fields.\n");
-	fprintf(ofp, "      --syscall           List available system calls.\n");
-#if 0
-	fprintf(ofp, "  -p, --pid PID           List user-space events by PID\n");
-#endif
-	fprintf(ofp, "\n");
-	fprintf(ofp, "Session Options:\n");
-	fprintf(ofp, "  -c, --channel NAME      List details of a channel\n");
-	fprintf(ofp, "  -d, --domain            List available domain(s)\n");
-	fprintf(ofp, "\n");
-}
-
 /*
  * Get command line from /proc for a specific pid.
  *
@@ -116,7 +78,8 @@ static char *get_cmdline_by_pid(pid_t pid)
 	int ret;
 	FILE *fp = NULL;
 	char *cmdline = NULL;
-	char path[20];	/* Can't go bigger than /proc/65535/cmdline */
+	/* Can't go bigger than /proc/LTTNG_MAX_PID/cmdline */
+	char path[sizeof("/proc//cmdline") + sizeof(LTTNG_MAX_PID_STR) - 1];
 
 	snprintf(path, sizeof(path), "/proc/%d/cmdline", pid);
 	fp = fopen(path, "r");
@@ -127,12 +90,12 @@ static char *get_cmdline_by_pid(pid_t pid)
 	/* Caller must free() *cmdline */
 	cmdline = zmalloc(PATH_MAX);
 	if (!cmdline) {
-		perror("malloc cmdline");
+		PERROR("malloc cmdline");
 		goto end;
 	}
 	ret = fread(cmdline, 1, PATH_MAX, fp);
 	if (ret < 0) {
-		perror("fread proc list");
+		PERROR("fread proc list");
 	}
 
 end:
@@ -175,21 +138,9 @@ const char *enabled_string(int value)
 }
 
 static
-const char *filter_string(int value)
+const char *safe_string(const char *str)
 {
-	switch (value) {
-	case 1:	return " [with filter]";
-	default: return "";
-	}
-}
-
-static
-const char *exclusion_string(int value)
-{
-	switch (value) {
-	case 1: return " [has exclusions]";
-	default: return "";
-	}
+	return str ? str : "";
 }
 
 static const char *logleveltype_string(enum lttng_loglevel_type value)
@@ -221,11 +172,104 @@ static const char *bitness_event(enum lttng_event_flag flags)
 	}
 }
 
+/*
+ * Get exclusion names message for a single event.
+ *
+ * Returned pointer must be freed by caller. Returns NULL on error.
+ */
+static char *get_exclusion_names_msg(struct lttng_event *event)
+{
+	int ret;
+	int exclusion_count;
+	char *exclusion_msg = NULL;
+	char *at;
+	size_t i;
+	const char * const exclusion_fmt = " [exclusions: ";
+	const size_t exclusion_fmt_len = strlen(exclusion_fmt);
+
+	exclusion_count = lttng_event_get_exclusion_name_count(event);
+	if (exclusion_count < 0) {
+		goto end;
+	} else if (exclusion_count == 0) {
+		/*
+		 * No exclusions: return copy of empty string so that
+		 * it can be freed by caller.
+		 */
+		exclusion_msg = strdup("");
+		goto end;
+	}
+
+	/*
+	 * exclusion_msg's size is bounded by the exclusion_fmt string,
+	 * a comma per entry, the entry count (fixed-size), a closing
+	 * bracket, and a trailing \0.
+	 */
+	exclusion_msg = malloc(exclusion_count +
+			exclusion_count * LTTNG_SYMBOL_NAME_LEN +
+			exclusion_fmt_len + 1);
+	if (!exclusion_msg) {
+		goto end;
+	}
+
+	at = strcpy(exclusion_msg, exclusion_fmt) + exclusion_fmt_len;
+	for (i = 0; i < exclusion_count; ++i) {
+		const char *name;
+
+		/* Append comma between exclusion names */
+		if (i > 0) {
+			*at = ',';
+			at++;
+		}
+
+		ret = lttng_event_get_exclusion_name(event, i, &name);
+		if (ret) {
+			/* Prints '?' on local error; should never happen */
+			*at = '?';
+			at++;
+			continue;
+		}
+
+		/* Append exclusion name */
+		at += sprintf(at, "%s", name);
+	}
+
+	/* This also puts a final '\0' at the end of exclusion_msg */
+	strcpy(at, "]");
+
+end:
+	return exclusion_msg;
+}
+
 /*
  * Pretty print single event.
  */
 static void print_events(struct lttng_event *event)
 {
+	int ret;
+	const char *filter_str;
+	char *filter_msg = NULL;
+	char *exclusion_msg = NULL;
+
+	ret = lttng_event_get_filter_expression(event, &filter_str);
+
+	if (ret) {
+		filter_msg = strdup(" [failed to retrieve filter]");
+	} else if (filter_str) {
+		const char * const filter_fmt = " [filter: '%s']";
+
+		filter_msg = malloc(strlen(filter_str) +
+				strlen(filter_fmt) + 1);
+		if (filter_msg) {
+			sprintf(filter_msg, filter_fmt,
+					filter_str);
+		}
+	}
+
+	exclusion_msg = get_exclusion_names_msg(event);
+	if (!exclusion_msg) {
+		exclusion_msg = strdup(" [failed to retrieve exclusions]");
+	}
+
 	switch (event->type) {
 	case LTTNG_EVENT_TRACEPOINT:
 	{
@@ -237,22 +281,22 @@ static void print_events(struct lttng_event *event)
 				mi_lttng_loglevel_string(event->loglevel, handle->domain.type),
 				event->loglevel,
 				enabled_string(event->enabled),
-				exclusion_string(event->exclusion),
-				filter_string(event->filter));
+				safe_string(exclusion_msg),
+				safe_string(filter_msg));
 		} else {
 			MSG("%s%s (type: tracepoint)%s%s%s",
 				indent6,
 				event->name,
 				enabled_string(event->enabled),
-				exclusion_string(event->exclusion),
-				filter_string(event->filter));
+				safe_string(exclusion_msg),
+				safe_string(filter_msg));
 		}
 		break;
 	}
 	case LTTNG_EVENT_FUNCTION:
 		MSG("%s%s (type: function)%s%s", indent6,
 				event->name, enabled_string(event->enabled),
-				filter_string(event->filter));
+				safe_string(filter_msg));
 		if (event->attr.probe.addr != 0) {
 			MSG("%saddr: 0x%" PRIx64, indent8, event->attr.probe.addr);
 		} else {
@@ -263,7 +307,7 @@ static void print_events(struct lttng_event *event)
 	case LTTNG_EVENT_PROBE:
 		MSG("%s%s (type: probe)%s%s", indent6,
 				event->name, enabled_string(event->enabled),
-				filter_string(event->filter));
+				safe_string(filter_msg));
 		if (event->attr.probe.addr != 0) {
 			MSG("%saddr: 0x%" PRIx64, indent8, event->attr.probe.addr);
 		} else {
@@ -274,25 +318,29 @@ static void print_events(struct lttng_event *event)
 	case LTTNG_EVENT_FUNCTION_ENTRY:
 		MSG("%s%s (type: function)%s%s", indent6,
 				event->name, enabled_string(event->enabled),
-				filter_string(event->filter));
+				safe_string(filter_msg));
 		MSG("%ssymbol: \"%s\"", indent8, event->attr.ftrace.symbol_name);
 		break;
 	case LTTNG_EVENT_SYSCALL:
-		MSG("%s%s%s%s%s", indent6, event->name,
+		MSG("%s%s%s%s%s%s", indent6, event->name,
 				(opt_syscall ? "" : " (type:syscall)"),
 				enabled_string(event->enabled),
-				bitness_event(event->flags));
+				bitness_event(event->flags),
+				safe_string(filter_msg));
 		break;
 	case LTTNG_EVENT_NOOP:
 		MSG("%s (type: noop)%s%s", indent6,
 				enabled_string(event->enabled),
-				filter_string(event->filter));
+				safe_string(filter_msg));
 		break;
 	case LTTNG_EVENT_ALL:
 		/* We should never have "all" events in list. */
 		assert(0);
 		break;
 	}
+
+	free(filter_msg);
+	free(exclusion_msg);
 }
 
 static const char *field_type(struct lttng_event_field *field)
@@ -348,7 +396,7 @@ static int mi_list_agent_ust_events(struct lttng_event *events, int count,
 		goto end;
 	}
 
-	/* Open pids element */
+	/* Open pids element element */
 	ret = mi_lttng_pids_open(writer);
 	if (ret) {
 		goto end;
@@ -357,7 +405,7 @@ static int mi_list_agent_ust_events(struct lttng_event *events, int count,
 	for (i = 0; i < count; i++) {
 		if (cur_pid != events[i].pid) {
 			if (pid_element_open) {
-				/* Close the previous events and  pid element */
+				/* Close the previous events and pid element */
 				ret = mi_lttng_close_multi_element(writer, 2);
 				if (ret) {
 					goto end;
@@ -416,8 +464,8 @@ static int list_agent_events(void)
 {
 	int i, size, ret = CMD_SUCCESS;
 	struct lttng_domain domain;
-	struct lttng_handle *handle;
-	struct lttng_event *event_list;
+	struct lttng_handle *handle = NULL;
+	struct lttng_event *event_list = NULL;
 	pid_t cur_pid = 0;
 	char *cmdline = NULL;
 	const char *agent_domain_str;
@@ -427,6 +475,12 @@ static int list_agent_events(void)
 		domain.type = LTTNG_DOMAIN_JUL;
 	} else if (opt_log4j) {
 		domain.type = LTTNG_DOMAIN_LOG4J;
+	} else if (opt_python) {
+		domain.type = LTTNG_DOMAIN_PYTHON;
+	} else {
+		ERR("Invalid agent domain selected.");
+		ret = CMD_ERROR;
+		goto error;
 	}
 
 	agent_domain_str = get_domain_str(domain.type);
@@ -495,7 +549,7 @@ static int list_ust_events(void)
 	int i, size, ret = CMD_SUCCESS;
 	struct lttng_domain domain;
 	struct lttng_handle *handle;
-	struct lttng_event *event_list;
+	struct lttng_event *event_list = NULL;
 	pid_t cur_pid = 0;
 	char *cmdline = NULL;
 
@@ -567,6 +621,8 @@ static int mi_list_ust_event_fields(struct lttng_event_field *fields, int count,
 	int event_element_open = 0;
 	struct lttng_event cur_event;
 
+	memset(&cur_event, 0, sizeof(cur_event));
+
 	/* Open domains element */
 	ret = mi_lttng_domains_open(writer);
 	if (ret) {
@@ -589,7 +645,6 @@ static int mi_list_ust_event_fields(struct lttng_event_field *fields, int count,
 		if (cur_pid != fields[i].event.pid) {
 			if (pid_element_open) {
 				if (event_element_open) {
-
 					/* Close the previous field element and event. */
 					ret = mi_lttng_close_multi_element(writer, 2);
 					if (ret) {
@@ -663,7 +718,7 @@ static int mi_list_ust_event_fields(struct lttng_event_field *fields, int count,
 		}
 	}
 
-	/* Close pids, domain, domains */
+	/* Close pid, domain, domains */
 	ret = mi_lttng_close_multi_element(writer, 3);
 end:
 	return ret;
@@ -987,11 +1042,43 @@ static int list_session_agent_events(void)
 		}
 
 		for (i = 0; i < count; i++) {
-			MSG("%s- %s%s (loglevel%s %s)", indent4, events[i].name,
-					enabled_string(events[i].enabled),
-					logleveltype_string(events[i].loglevel_type),
-					mi_lttng_loglevel_string(events[i].loglevel,
-						handle->domain.type));
+			const char *filter_str;
+			char *filter_msg = NULL;
+			struct lttng_event *event = &events[i];
+
+			ret = lttng_event_get_filter_expression(event,
+					&filter_str);
+			if (ret) {
+				filter_msg = strdup(" [failed to retrieve filter]");
+			} else if (filter_str) {
+				const char * const filter_fmt =
+						" [filter: '%s']";
+
+				filter_msg = malloc(strlen(filter_str) +
+						strlen(filter_fmt) + 1);
+				if (filter_msg) {
+					sprintf(filter_msg, filter_fmt,
+							filter_str);
+				}
+			}
+
+			if (event->loglevel_type !=
+					LTTNG_EVENT_LOGLEVEL_ALL) {
+				MSG("%s- %s%s (loglevel%s %s)%s", indent4,
+						event->name,
+						enabled_string(event->enabled),
+						logleveltype_string(
+							event->loglevel_type),
+						mi_lttng_loglevel_string(
+							event->loglevel,
+							handle->domain.type),
+						safe_string(filter_msg));
+			} else {
+				MSG("%s- %s%s%s", indent4, event->name,
+						enabled_string(event->enabled),
+						safe_string(filter_msg));
+			}
+			free(filter_msg);
 		}
 
 		MSG("");
@@ -1078,16 +1165,35 @@ error:
  */
 static void print_channel(struct lttng_channel *channel)
 {
+	int ret;
+	uint64_t discarded_events, lost_packets;
+
+	ret = lttng_channel_get_discarded_event_count(channel,
+			&discarded_events);
+	if (ret) {
+		ERR("Failed to retrieve discarded event count of channel");
+		return;
+	}
+
+	ret = lttng_channel_get_lost_packet_count(channel,
+			&lost_packets);
+	if (ret) {
+		ERR("Failed to retrieve lost packet count of channel");
+		return;
+	}
+
 	MSG("- %s:%s\n", channel->name, enabled_string(channel->enabled));
 
 	MSG("%sAttributes:", indent4);
 	MSG("%soverwrite mode: %d", indent6, channel->attr.overwrite);
-	MSG("%ssubbufers size: %" PRIu64, indent6, channel->attr.subbuf_size);
-	MSG("%snumber of subbufers: %" PRIu64, indent6, channel->attr.num_subbuf);
+	MSG("%ssubbuffers size: %" PRIu64, indent6, channel->attr.subbuf_size);
+	MSG("%snumber of subbuffers: %" PRIu64, indent6, channel->attr.num_subbuf);
 	MSG("%sswitch timer interval: %u", indent6, channel->attr.switch_timer_interval);
 	MSG("%sread timer interval: %u", indent6, channel->attr.read_timer_interval);
 	MSG("%strace file count: %" PRIu64, indent6, channel->attr.tracefile_count);
 	MSG("%strace file size (bytes): %" PRIu64, indent6, channel->attr.tracefile_size);
+	MSG("%sdiscarded events: %" PRIu64, indent6, discarded_events);
+	MSG("%slost packets: %" PRIu64, indent6, lost_packets);
 	switch (channel->attr.output) {
 		case LTTNG_EVENT_SPLICE:
 			MSG("%soutput: splice()", indent6);
@@ -1240,6 +1346,97 @@ error_channels:
 	return ret;
 }
 
+/*
+ * List tracker PID(s) of session and domain.
+ */
+static int list_tracker_pids(void)
+{
+	int ret = 0;
+	int enabled;
+	int *pids = NULL;
+	size_t nr_pids;
+
+	ret = lttng_list_tracker_pids(handle,
+		&enabled, &pids, &nr_pids);
+	if (ret) {
+		return ret;
+	}
+	if (enabled) {
+		int i;
+		_MSG("PID tracker: [");
+
+		/* Mi tracker_pid element*/
+		if (writer) {
+			/* Open tracker_pid and targets elements */
+			ret = mi_lttng_pid_tracker_open(writer);
+			if (ret) {
+				goto end;
+			}
+		}
+
+		for (i = 0; i < nr_pids; i++) {
+			if (i) {
+				_MSG(",");
+			}
+			_MSG(" %d", pids[i]);
+
+			/* Mi */
+			if (writer) {
+				ret = mi_lttng_pid_target(writer, pids[i], 0);
+				if (ret) {
+					goto end;
+				}
+			}
+		}
+		_MSG(" ]\n\n");
+
+		/* Mi close tracker_pid and targets */
+		if (writer) {
+			ret = mi_lttng_close_multi_element(writer,2);
+			if (ret) {
+				goto end;
+			}
+		}
+	}
+end:
+	free(pids);
+	return ret;
+
+}
+
+/*
+ * List all tracker of a domain
+ */
+static int list_trackers(void)
+{
+	int ret;
+
+	/* Trackers listing */
+	if (lttng_opt_mi) {
+		ret = mi_lttng_trackers_open(writer);
+		if (ret) {
+			goto end;
+		}
+	}
+
+	/* pid tracker */
+	ret = list_tracker_pids();
+	if (ret) {
+		goto end;
+	}
+
+	if (lttng_opt_mi) {
+		/* Close trackers element */
+		ret = mi_lttng_writer_close_element(writer);
+		if (ret) {
+			goto end;
+		}
+	}
+
+end:
+	return ret;
+}
+
 /*
  * Machine interface
  * Find the session with session_name as name
@@ -1372,8 +1569,11 @@ static int list_sessions(const char *session_name)
 						active_string(sessions[i].enabled),
 						snapshot_string(sessions[i].snapshot_mode));
 				MSG("%sTrace path: %s", indent4, sessions[i].path);
-				MSG("%sLive timer interval (usec): %u\n", indent4,
-						sessions[i].live_timer_interval);
+				if (sessions[i].live_timer_interval != 0) {
+					MSG("%sLive timer interval (usec): %u", indent4,
+							sessions[i].live_timer_interval);
+				}
+				MSG("");
 			}
 		}
 
@@ -1469,6 +1669,9 @@ static int list_domains(const char *session_name)
 			case LTTNG_DOMAIN_LOG4J:
 				MSG("  - LOG4j (Logging for Java)");
 				break;
+			case LTTNG_DOMAIN_PYTHON:
+				MSG("  - Python (logging)");
+				break;
 			default:
 				break;
 			}
@@ -1496,7 +1699,6 @@ int cmd_list(int argc, const char **argv)
 	memset(&domain, 0, sizeof(domain));
 
 	if (argc < 1) {
-		usage(stderr);
 		ret = CMD_ERROR;
 		goto end;
 	}
@@ -1507,7 +1709,7 @@ int cmd_list(int argc, const char **argv)
 	while ((opt = poptGetNextOpt(pc)) != -1) {
 		switch (opt) {
 		case OPT_HELP:
-			usage(stdout);
+			SHOW_HELP();
 			goto end;
 		case OPT_USERSPACE:
 			opt_userspace = 1;
@@ -1516,7 +1718,6 @@ int cmd_list(int argc, const char **argv)
 			list_cmd_options(stdout, long_options);
 			goto end;
 		default:
-			usage(stderr);
 			ret = CMD_UNDEFINED;
 			goto end;
 		}
@@ -1561,6 +1762,8 @@ int cmd_list(int argc, const char **argv)
 		domain.type = LTTNG_DOMAIN_JUL;
 	} else if (opt_log4j) {
 		domain.type = LTTNG_DOMAIN_LOG4J;
+	} else if (opt_python) {
+		domain.type = LTTNG_DOMAIN_PYTHON;
 	}
 
 	if (!opt_kernel && opt_syscall) {
@@ -1569,7 +1772,7 @@ int cmd_list(int argc, const char **argv)
 		goto end;
 	}
 
-	if (opt_kernel || opt_userspace || opt_jul || opt_log4j) {
+	if (opt_kernel || opt_userspace || opt_jul || opt_log4j || opt_python) {
 		handle = lttng_create_handle(session_name, &domain);
 		if (handle == NULL) {
 			ret = CMD_FATAL;
@@ -1578,7 +1781,8 @@ int cmd_list(int argc, const char **argv)
 	}
 
 	if (session_name == NULL) {
-		if (!opt_kernel && !opt_userspace && !opt_jul && !opt_log4j) {
+		if (!opt_kernel && !opt_userspace && !opt_jul && !opt_log4j
+				&& !opt_python) {
 			ret = list_sessions(NULL);
 			if (ret) {
 				goto end;
@@ -1607,7 +1811,7 @@ int cmd_list(int argc, const char **argv)
 				goto end;
 			}
 		}
-		if (opt_jul || opt_log4j) {
+		if (opt_jul || opt_log4j || opt_python) {
 			ret = list_agent_events();
 			if (ret) {
 				goto end;
@@ -1655,6 +1859,14 @@ int cmd_list(int argc, const char **argv)
 
 			}
 
+
+			/* Trackers */
+			ret = list_trackers();
+			if (ret) {
+				goto end;
+			}
+
+			/* Channels */
 			ret = list_channels(opt_channel);
 			if (ret) {
 				goto end;
@@ -1705,6 +1917,9 @@ int cmd_list(int argc, const char **argv)
 				case LTTNG_DOMAIN_LOG4J:
 					MSG("=== Domain: LOG4j (Logging for Java) ===\n");
 					break;
+				case LTTNG_DOMAIN_PYTHON:
+					MSG("=== Domain: Python (logging) ===\n");
+					break;
 				default:
 					MSG("=== Domain: Unimplemented ===\n");
 					break;
@@ -1730,12 +1945,26 @@ int cmd_list(int argc, const char **argv)
 				}
 
 				if (domains[i].type == LTTNG_DOMAIN_JUL ||
-						domains[i].type == LTTNG_DOMAIN_LOG4J) {
+						domains[i].type == LTTNG_DOMAIN_LOG4J ||
+						domains[i].type == LTTNG_DOMAIN_PYTHON) {
 					ret = list_session_agent_events();
 					if (ret) {
 						goto end;
 					}
-					continue;
+
+					goto next_domain;
+				}
+
+				switch (domains[i].type) {
+				case LTTNG_DOMAIN_KERNEL:
+				case LTTNG_DOMAIN_UST:
+					ret = list_trackers();
+					if (ret) {
+						goto end;
+					}
+					break;
+				default:
+					break;
 				}
 
 				ret = list_channels(opt_channel);
@@ -1743,6 +1972,7 @@ int cmd_list(int argc, const char **argv)
 					goto end;
 				}
 
+next_domain:
 				if (lttng_opt_mi) {
 					/* Close domain element */
 					ret = mi_lttng_writer_close_element(writer);
