@@ -1,6 +1,7 @@
 /*
  * Copyright (C) 2014 - Jonathan Rajotte <jonathan.r.julien@gmail.com>
  *                    - Olivier Cotte <olivier.cotte@polymtl.ca>
+ * Copyright (C) 2016 - Jérémie Galarneau <jeremie.galarneau@efficios.com>
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License, version 2 only, as
@@ -16,19 +17,36 @@
  * Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
-
-#include <include/config.h>
-#include <common/config/config.h>
+#define _LGPL_SOURCE
+#include <common/config/session-config.h>
+#include <common/defaults.h>
 #include <lttng/snapshot-internal.h>
+#include <lttng/channel.h>
 #include "mi-lttng.h"
 
 #include <assert.h>
 
+#define MI_SCHEMA_MAJOR_VERSION 3
+#define MI_SCHEMA_MINOR_VERSION 0
+
+/* Machine interface namespace URI */
+LTTNG_HIDDEN const char * const mi_lttng_xmlns = "xmlns";
+LTTNG_HIDDEN const char * const mi_lttng_xmlns_xsi = "xmlns:xsi";
+LTTNG_HIDDEN const char * const mi_lttng_w3_schema_uri = "http://www.w3.org/2001/XMLSchema-instance";
+LTTNG_HIDDEN const char * const mi_lttng_schema_location = "xsi:schemaLocation";
+LTTNG_HIDDEN const char * const mi_lttng_schema_location_uri =
+	DEFAULT_LTTNG_MI_NAMESPACE " "
+	"http://lttng.org/xml/schemas/lttng-mi/" XSTR(MI_SCHEMA_MAJOR_VERSION)
+	"/lttng-mi-" XSTR(MI_SCHEMA_MAJOR_VERSION) "."
+	XSTR(MI_SCHEMA_MINOR_VERSION) ".xsd";
+LTTNG_HIDDEN const char * const mi_lttng_schema_version = "schemaVersion";
+LTTNG_HIDDEN const char * const mi_lttng_schema_version_value = XSTR(MI_SCHEMA_MAJOR_VERSION)
+	"." XSTR(MI_SCHEMA_MINOR_VERSION);
+
 /* Strings related to command */
 const char * const mi_lttng_element_command = "command";
 const char * const mi_lttng_element_command_action = "snapshot_action";
 const char * const mi_lttng_element_command_add_context = "add-context";
-const char * const mi_lttng_element_command_calibrate = "calibrate";
 const char * const mi_lttng_element_command_create = "create";
 const char * const mi_lttng_element_command_destroy = "destroy";
 const char * const mi_lttng_element_command_disable_channel = "disable-channel";
@@ -37,6 +55,10 @@ const char * const mi_lttng_element_command_enable_channels = "enable-channel";
 const char * const mi_lttng_element_command_enable_event = "enable-event";
 const char * const mi_lttng_element_command_list = "list";
 const char * const mi_lttng_element_command_load = "load";
+LTTNG_HIDDEN const char * const mi_lttng_element_command_metadata = "metadata";
+LTTNG_HIDDEN const char * const mi_lttng_element_command_metadata_action = "metadata_action";
+LTTNG_HIDDEN const char * const mi_lttng_element_command_regenerate = "regenerate";
+LTTNG_HIDDEN const char * const mi_lttng_element_command_regenerate_action = "regenerate_action";
 const char * const mi_lttng_element_command_name = "name";
 const char * const mi_lttng_element_command_output = "output";
 const char * const mi_lttng_element_command_save = "save";
@@ -49,6 +71,8 @@ const char * const mi_lttng_element_command_snapshot_record = "record_snapshot";
 const char * const mi_lttng_element_command_start = "start";
 const char * const mi_lttng_element_command_stop = "stop";
 const char * const mi_lttng_element_command_success = "success";
+const char * const mi_lttng_element_command_track = "track";
+const char * const mi_lttng_element_command_untrack = "untrack";
 const char * const mi_lttng_element_command_version = "version";
 
 /* Strings related to version command */
@@ -66,17 +90,10 @@ const char * const mi_lttng_element_version_web = "url";
 const char * const mi_lttng_element_event_field = "event_field";
 const char * const mi_lttng_element_event_fields = "event_fields";
 
-/* String related to lttng_event_context */
-const char * const mi_lttng_context_type_perf_counter = "PERF_COUNTER";
-const char * const mi_lttng_context_type_perf_cpu_counter = "PERF_CPU_COUNTER";
-const char * const mi_lttng_context_type_perf_thread_counter = "PERF_THREAD_COUNTER";
-
 /* String related to lttng_event_perf_counter_ctx */
-const char * const mi_lttng_element_perf_counter_context = "perf_counter_context";
+const char * const mi_lttng_element_perf_counter_context = "perf";
 
 /* Strings related to pid */
-const char * const mi_lttng_element_pids = "pids";
-const char * const mi_lttng_element_pid = "pid";
 const char * const mi_lttng_element_pid_id = "id";
 
 /* Strings related to save command */
@@ -84,6 +101,8 @@ const char * const mi_lttng_element_save = "save";
 
 /* Strings related to load command */
 const char * const mi_lttng_element_load = "load";
+LTTNG_HIDDEN const char * const mi_lttng_element_load_overrides = "overrides";
+LTTNG_HIDDEN const char * const mi_lttng_element_load_override_url = "url";
 
 /* General elements of mi_lttng */
 const char * const mi_lttng_element_empty = "";
@@ -135,16 +154,20 @@ const char * const mi_lttng_loglevel_str_log4j_debug = "LOG4J_DEBUG";
 const char * const mi_lttng_loglevel_str_log4j_trace = "LOG4J_TRACE";
 const char * const mi_lttng_loglevel_str_log4j_all = "LOG4J_ALL";
 
+/* String related to loglevel Python */
+const char * const mi_lttng_loglevel_str_python_critical = "PYTHON_CRITICAL";
+const char * const mi_lttng_loglevel_str_python_error = "PYTHON_ERROR";
+const char * const mi_lttng_loglevel_str_python_warning = "PYTHON_WARNING";
+const char * const mi_lttng_loglevel_str_python_info = "PYTHON_INFO";
+const char * const mi_lttng_loglevel_str_python_debug = "PYTHON_DEBUG";
+const char * const mi_lttng_loglevel_str_python_notset = "PYTHON_NOTSET";
+
 /* String related to loglevel type */
 const char * const mi_lttng_loglevel_type_all = "ALL";
 const char * const mi_lttng_loglevel_type_range = "RANGE";
 const char * const mi_lttng_loglevel_type_single = "SINGLE";
 const char * const mi_lttng_loglevel_type_unknown = "UNKNOWN";
 
-/* String related to lttng_calibrate */
-const char * const mi_lttng_element_calibrate = "calibrate";
-const char * const mi_lttng_element_calibrate_function = "FUNCTION";
-
 /* String related to a lttng_snapshot_output */
 const char * const mi_lttng_element_snapshot_ctrl_url = "ctrl_url";
 const char * const mi_lttng_element_snapshot_data_url = "data_url";
@@ -153,6 +176,19 @@ const char * const mi_lttng_element_snapshot_n_ptr = "n_ptr";
 const char * const mi_lttng_element_snapshot_session_name = "session_name";
 const char * const mi_lttng_element_snapshots = "snapshots";
 
+/* String related to track/untrack command */
+const char * const mi_lttng_element_track_untrack_all_wildcard = "*";
+
+/* Deprecated symbols preserved for ABI compatibility. */
+const char * const mi_lttng_context_type_perf_counter;
+const char * const mi_lttng_context_type_perf_cpu_counter;
+const char * const mi_lttng_context_type_perf_thread_counter;
+const char * const mi_lttng_element_track_untrack_pid_target;
+const char * const mi_lttng_element_track_untrack_targets;
+const char * const mi_lttng_element_calibrate;
+const char * const mi_lttng_element_calibrate_function;
+const char * const mi_lttng_element_command_calibrate;
+
 /* This is a merge of jul loglevel and regular loglevel
  * Those should never overlap by definition
  * (see struct lttng_event loglevel)
@@ -250,10 +286,27 @@ const char *mi_lttng_loglevel_string(int value, enum lttng_domain_type domain)
 			return mi_lttng_loglevel_str_unknown;
 		}
 		break;
+	case LTTNG_DOMAIN_PYTHON:
+		switch (value) {
+		case LTTNG_LOGLEVEL_PYTHON_CRITICAL:
+			return mi_lttng_loglevel_str_python_critical;
+		case LTTNG_LOGLEVEL_PYTHON_ERROR:
+			return mi_lttng_loglevel_str_python_error;
+		case LTTNG_LOGLEVEL_PYTHON_WARNING:
+			return mi_lttng_loglevel_str_python_warning;
+		case LTTNG_LOGLEVEL_PYTHON_INFO:
+			return mi_lttng_loglevel_str_python_info;
+		case LTTNG_LOGLEVEL_PYTHON_DEBUG:
+			return mi_lttng_loglevel_str_python_debug;
+		case LTTNG_LOGLEVEL_PYTHON_NOTSET:
+			return mi_lttng_loglevel_str_python_notset;
+		default:
+			return mi_lttng_loglevel_str_unknown;
+		}
+		break;
+	default:
+		return mi_lttng_loglevel_str_unknown;
 	}
-
-	/* Reaching this means the domain is unknown. */
-	return mi_lttng_loglevel_str_unknown;
 }
 
 LTTNG_HIDDEN
@@ -294,18 +347,12 @@ const char *mi_lttng_eventtype_string(enum lttng_event_type value)
 	}
 }
 
-LTTNG_HIDDEN
+static
 const char *mi_lttng_event_contexttype_string(enum lttng_event_context_type val)
 {
 	switch (val) {
 	case LTTNG_EVENT_CONTEXT_PID:
 		return config_event_context_pid;
-	case LTTNG_EVENT_CONTEXT_PERF_COUNTER:
-		return mi_lttng_context_type_perf_counter;
-	case LTTNG_EVENT_CONTEXT_PERF_THREAD_COUNTER:
-		return mi_lttng_context_type_perf_thread_counter;
-	case LTTNG_EVENT_CONTEXT_PERF_CPU_COUNTER:
-		return mi_lttng_context_type_perf_cpu_counter;
 	case LTTNG_EVENT_CONTEXT_PROCNAME:
 		return config_event_context_procname;
 	case LTTNG_EVENT_CONTEXT_PRIO:
@@ -363,9 +410,12 @@ const char *mi_lttng_domaintype_string(enum lttng_domain_type value)
 		return config_domain_type_jul;
 	case LTTNG_DOMAIN_LOG4J:
 		return config_domain_type_log4j;
+	case LTTNG_DOMAIN_PYTHON:
+		return config_domain_type_python;
 	default:
 		/* Should not have an unknown domain */
 		assert(0);
+		return NULL;
 	}
 }
 
@@ -382,25 +432,10 @@ const char *mi_lttng_buffertype_string(enum lttng_buffer_type value)
 	default:
 		/* Should not have an unknow buffer type */
 		assert(0);
+		return NULL;
 	}
 }
 
-LTTNG_HIDDEN
-const char *mi_lttng_calibratetype_string(enum lttng_calibrate_type val)
-{
-	const char *ret;
-
-	switch (val) {
-	case LTTNG_CALIBRATE_FUNCTION:
-		ret = mi_lttng_element_calibrate_function;
-		break;
-	default:
-		ret = mi_lttng_element_empty;
-		break;
-	}
-	return ret;
-}
-
 LTTNG_HIDDEN
 struct mi_writer *mi_lttng_writer_create(int fd_output, int mi_output_type)
 {
@@ -454,10 +489,42 @@ int mi_lttng_writer_command_open(struct mi_writer *writer, const char *command)
 {
 	int ret;
 
-	ret = mi_lttng_writer_open_element(writer, mi_lttng_element_command);
+	/*
+	 * A command is always the MI's root node, it must declare the current
+	 * namespace and schema URIs and the schema's version.
+	 */
+	ret = config_writer_open_element(writer->writer,
+			mi_lttng_element_command);
 	if (ret) {
 		goto end;
 	}
+
+	ret = config_writer_write_attribute(writer->writer,
+			mi_lttng_xmlns, DEFAULT_LTTNG_MI_NAMESPACE);
+	if (ret) {
+		goto end;
+	}
+
+	ret = config_writer_write_attribute(writer->writer,
+			mi_lttng_xmlns_xsi, mi_lttng_w3_schema_uri);
+	if (ret) {
+		goto end;
+	}
+
+	ret = config_writer_write_attribute(writer->writer,
+			mi_lttng_schema_location,
+			mi_lttng_schema_location_uri);
+	if (ret) {
+		goto end;
+	}
+
+	ret = config_writer_write_attribute(writer->writer,
+			mi_lttng_schema_version,
+			mi_lttng_schema_version_value);
+	if (ret) {
+		goto end;
+	}
+
 	ret = mi_lttng_writer_write_element_string(writer,
 			mi_lttng_element_command_name, command);
 end:
@@ -793,9 +860,22 @@ int mi_lttng_channel_attr(struct mi_writer *writer,
 		struct lttng_channel_attr *attr)
 {
 	int ret = 0;
+	struct lttng_channel *chan = caa_container_of(attr,
+			struct lttng_channel, attr);
+	uint64_t discarded_events, lost_packets;
 
 	assert(attr);
 
+	ret = lttng_channel_get_discarded_event_count(chan, &discarded_events);
+	if (ret) {
+		goto end;
+	}
+
+	ret = lttng_channel_get_lost_packet_count(chan, &lost_packets);
+	if (ret) {
+		goto end;
+	}
+
 	/* Opening Attributes */
 	ret = mi_lttng_writer_open_element(writer, config_element_attributes);
 	if (ret) {
@@ -874,6 +954,22 @@ int mi_lttng_channel_attr(struct mi_writer *writer,
 		goto end;
 	}
 
+	/* Discarded events */
+	ret = mi_lttng_writer_write_element_unsigned_int(writer,
+		config_element_discarded_events,
+		discarded_events);
+	if (ret) {
+		goto end;
+	}
+
+	/* Lost packets */
+	ret = mi_lttng_writer_write_element_unsigned_int(writer,
+		config_element_lost_packets,
+		lost_packets);
+	if (ret) {
+		goto end;
+	}
+
 	/* Closing attributes */
 	ret = mi_lttng_writer_close_element(writer);
 	if (ret) {
@@ -889,6 +985,7 @@ int mi_lttng_event_common_attributes(struct mi_writer *writer,
 		struct lttng_event *event)
 {
 	int ret;
+	const char *filter_expression;
 
 	/* Open event element */
 	ret = mi_lttng_writer_open_element(writer, config_element_event);
@@ -917,9 +1014,65 @@ int mi_lttng_event_common_attributes(struct mi_writer *writer,
 		goto end;
 	}
 
-	/* Event filter enabled? */
-	ret = mi_lttng_writer_write_element_bool(writer,
-			config_element_filter, event->filter);
+	/* Event filter expression */
+	ret = lttng_event_get_filter_expression(event, &filter_expression);
+	if (ret) {
+		goto end;
+	}
+
+	if (filter_expression) {
+		ret = mi_lttng_writer_write_element_string(writer,
+				config_element_filter_expression,
+				filter_expression);
+		if (ret) {
+			goto end;
+		}
+	}
+
+end:
+	return ret;
+}
+
+static int write_event_exclusions(struct mi_writer *writer,
+		struct lttng_event *event)
+{
+	int i;
+	int ret;
+	int exclusion_count;
+
+	/* Open event exclusions */
+	ret = mi_lttng_writer_open_element(writer, config_element_exclusions);
+	if (ret) {
+		goto end;
+	}
+
+	exclusion_count = lttng_event_get_exclusion_name_count(event);
+	if (exclusion_count < 0) {
+		ret = exclusion_count;
+		goto end;
+	}
+
+	for (i = 0; i < exclusion_count; i++) {
+		const char *name;
+
+		ret = lttng_event_get_exclusion_name(event, i, &name);
+		if (ret) {
+			/* Close exclusions */
+			mi_lttng_writer_close_element(writer);
+			goto end;
+		}
+
+		ret = mi_lttng_writer_write_element_string(writer,
+				config_element_exclusion, name);
+		if (ret) {
+			/* Close exclusions */
+			mi_lttng_writer_close_element(writer);
+			goto end;
+		}
+	}
+
+	/* Close exclusions */
+	ret = mi_lttng_writer_close_element(writer);
 
 end:
 	return ret;
@@ -947,12 +1100,8 @@ int mi_lttng_event_tracepoint_loglevel(struct mi_writer *writer,
 		goto end;
 	}
 
-	/* event exclusion filter */
-	ret = mi_lttng_writer_write_element_bool(writer,
-			config_element_exclusion, event->exclusion);
-	if (ret) {
-		goto end;
-	}
+	/* Event exclusions */
+	ret = write_event_exclusions(writer, event);
 
 end:
 	return ret;
@@ -963,8 +1112,7 @@ int mi_lttng_event_tracepoint_no_loglevel(struct mi_writer *writer,
 		struct lttng_event *event)
 {
 	/* event exclusion filter */
-	return mi_lttng_writer_write_element_bool(writer,
-			config_element_exclusion, event->exclusion);
+	return write_event_exclusions(writer, event);
 }
 
 LTTNG_HIDDEN
@@ -1091,19 +1239,46 @@ end:
 }
 
 LTTNG_HIDDEN
-int mi_lttng_pids_open(struct mi_writer *writer)
+int mi_lttng_trackers_open(struct mi_writer *writer)
 {
-	return mi_lttng_writer_open_element(writer, mi_lttng_element_pids);
+	return mi_lttng_writer_open_element(writer, config_element_trackers);
 }
 
 LTTNG_HIDDEN
-int mi_lttng_pid(struct mi_writer *writer, pid_t pid , const char *cmdline,
+int mi_lttng_pid_tracker_open(struct mi_writer *writer)
+{
+	int ret;
+
+	/* Open element pid_tracker */
+	ret = mi_lttng_writer_open_element(writer, config_element_pid_tracker);
+	if (ret) {
+		goto end;
+	}
+
+	/* Open targets element */
+	ret = mi_lttng_targets_open(writer);
+end:
+	return ret;
+}
+
+LTTNG_HIDDEN
+int mi_lttng_pids_open(struct mi_writer *writer)
+{
+	return mi_lttng_writer_open_element(writer, config_element_pids);
+}
+
+/*
+ * TODO: move the listing of pid for user agent to process semantic on
+ * mi api bump. The use of process element break the mi api.
+ */
+LTTNG_HIDDEN
+int mi_lttng_pid(struct mi_writer *writer, pid_t pid , const char *name,
 		int is_open)
 {
 	int ret;
 
-	/* Open element pid */
-	ret = mi_lttng_writer_open_element(writer, mi_lttng_element_pid);
+	/* Open pid process */
+	ret = mi_lttng_writer_open_element(writer, config_element_pid);
 	if (ret) {
 		goto end;
 	}
@@ -1116,10 +1291,12 @@ int mi_lttng_pid(struct mi_writer *writer, pid_t pid , const char *cmdline,
 	}
 
 	/* Writing name of the process */
-	ret = mi_lttng_writer_write_element_string(writer, config_element_name,
-			cmdline);
-	if (ret) {
-		goto end;
+	if (name) {
+		ret = mi_lttng_writer_write_element_string(writer, config_element_name,
+				name);
+		if (ret) {
+			goto end;
+		}
 	}
 
 	if (!is_open) {
@@ -1131,6 +1308,51 @@ end:
 	return ret;
 }
 
+LTTNG_HIDDEN
+int mi_lttng_targets_open(struct mi_writer *writer)
+{
+	return mi_lttng_writer_open_element(writer,
+			config_element_targets);
+}
+
+LTTNG_HIDDEN
+int mi_lttng_pid_target(struct mi_writer *writer, pid_t pid, int is_open)
+{
+	int ret;
+
+	ret = mi_lttng_writer_open_element(writer,
+			config_element_target_pid);
+	if (ret) {
+		goto end;
+	}
+
+	/* Writing pid number
+	 * Special case for element all on track untrack command
+	 * All pid is represented as wildcard *
+	 */
+	if ((int) pid == -1) {
+		ret = mi_lttng_writer_write_element_string(writer,
+				config_element_pid,
+				mi_lttng_element_track_untrack_all_wildcard);
+	} else {
+		ret = mi_lttng_writer_write_element_signed_int(writer,
+				config_element_pid, (int) pid);
+	}
+	if (ret) {
+		goto end;
+	}
+
+	if (!is_open) {
+		ret = mi_lttng_writer_close_element(writer);
+		if (ret) {
+			goto end;
+		}
+	}
+
+end:
+	return ret;
+}
+
 LTTNG_HIDDEN
 int mi_lttng_event_fields_open(struct mi_writer *writer)
 {
@@ -1187,78 +1409,6 @@ end:
 	return ret;
 }
 
-LTTNG_HIDDEN
-int mi_lttng_calibrate(struct mi_writer *writer,
-		struct lttng_calibrate *calibrate)
-{
-	int ret;
-
-	/* Open calibrate element */
-	ret = mi_lttng_writer_open_element(writer, mi_lttng_element_calibrate);
-	if (ret) {
-		goto end;
-	}
-
-	/* Calibration type */
-	ret = mi_lttng_writer_write_element_string(writer, config_element_type,
-			mi_lttng_calibratetype_string(calibrate->type));
-	if (ret) {
-		goto end;
-	}
-
-	/* Closing calibrate element */
-	ret = mi_lttng_writer_close_element(writer);
-end:
-	return ret;
-}
-LTTNG_HIDDEN
-int mi_lttng_context(struct mi_writer *writer,
-		struct lttng_event_context *context, int is_open)
-{
-	int ret;
-	const char *type_string;
-	struct lttng_event_perf_counter_ctx *perf_context;
-	/* Open context */
-	ret = mi_lttng_writer_open_element(writer , config_element_context);
-	if (ret) {
-		goto end;
-	}
-
-	type_string = mi_lttng_event_contexttype_string(context->ctx);
-	if (!type_string) {
-		ret = -LTTNG_ERR_INVALID;
-		goto end;
-	}
-
-	/* Print context type */
-	ret = mi_lttng_writer_write_element_string(writer, config_element_type,
-			type_string);
-
-	/* Special case for PERF_*_COUNTER
-	 * print the lttng_event_perf_counter_ctx*/
-	switch (context->ctx) {
-	case LTTNG_EVENT_CONTEXT_PERF_COUNTER:
-	case LTTNG_EVENT_CONTEXT_PERF_THREAD_COUNTER:
-	case LTTNG_EVENT_CONTEXT_PERF_CPU_COUNTER:
-		perf_context = &context->u.perf_counter;
-		ret =  mi_lttng_perf_counter_context(writer, perf_context);
-		if (ret) {
-			goto end;
-		}
-		break;
-	default:
-		break;
-	}
-
-	/* Close context */
-	if (!is_open) {
-		ret = mi_lttng_writer_close_element(writer);
-	}
-
-end:
-	return ret;
-}
-
 LTTNG_HIDDEN
 int mi_lttng_perf_counter_context(struct mi_writer *writer,
 		struct lttng_event_perf_counter_ctx  *perf_context)
@@ -1299,6 +1449,102 @@ end:
 	return ret;
 }
 
+static
+int mi_lttng_app_context(struct mi_writer *writer,
+		const char *provider_name, const char *ctx_name)
+{
+	int ret;
+
+	/* Open app */
+	ret = mi_lttng_writer_open_element(writer,
+			config_element_context_app);
+	if (ret) {
+		goto end;
+	}
+
+	/* provider_name */
+	ret = mi_lttng_writer_write_element_string(writer,
+			config_element_context_app_provider_name,
+			provider_name);
+	if (ret) {
+		goto end;
+	}
+
+	/* ctx_name */
+	ret = mi_lttng_writer_write_element_string(writer,
+			config_element_context_app_ctx_name, ctx_name);
+	if (ret) {
+		goto end;
+	}
+
+	/* Close app */
+	ret = mi_lttng_writer_close_element(writer);
+end:
+	return ret;
+}
+
+LTTNG_HIDDEN
+int mi_lttng_context(struct mi_writer *writer,
+		struct lttng_event_context *context, int is_open)
+{
+	int ret;
+
+	/* Open context */
+	ret = mi_lttng_writer_open_element(writer , config_element_context);
+	if (ret) {
+		goto end;
+	}
+
+	/* Special case for PERF_*_COUNTER
+	 * print the lttng_event_perf_counter_ctx*/
+	switch (context->ctx) {
+	case LTTNG_EVENT_CONTEXT_PERF_COUNTER:
+	case LTTNG_EVENT_CONTEXT_PERF_THREAD_COUNTER:
+	case LTTNG_EVENT_CONTEXT_PERF_CPU_COUNTER:
+	{
+		struct lttng_event_perf_counter_ctx *perf_context =
+				&context->u.perf_counter;
+		ret =  mi_lttng_perf_counter_context(writer, perf_context);
+		if (ret) {
+			goto end;
+		}
+		break;
+	}
+	case LTTNG_EVENT_CONTEXT_APP_CONTEXT:
+	{
+		ret = mi_lttng_app_context(writer,
+				context->u.app_ctx.provider_name,
+				context->u.app_ctx.ctx_name);
+		if (ret) {
+			goto end;
+		}
+		break;
+	}
+	default:
+	{
+		const char *type_string = mi_lttng_event_contexttype_string(
+				context->ctx);
+		if (!type_string) {
+			ret = -LTTNG_ERR_INVALID;
+			goto end;
+		}
+
+		/* Print context type */
+		ret = mi_lttng_writer_write_element_string(writer,
+				config_element_type, type_string);
+		break;
+	}
+	}
+
+	/* Close context */
+	if (!is_open) {
+		ret = mi_lttng_writer_close_element(writer);
+	}
+
+end:
+	return ret;
+}
+
 LTTNG_HIDDEN
 int mi_lttng_snapshot_output_session_name(struct mi_writer *writer,
 		const char *session_name)
